// StegoWav - a tool for steganography on Microsoft wave files
//
// Copyright (C) 1996 by Peter Heist <heistp@rpi.edu>.  All rights reserved.
//
// THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
// ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
// IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
// ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE
// FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
// DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
// OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
// HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
// LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
// OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
// SUCH DAMAGE.

import java.lang.*;
import java.io.*;
import java.util.*;
import java.crypt.*;
import WavFile;

public class AudioCover 
    {
    // hidden file header information
    final byte[] magicArray = { (byte) 0x05, (byte) 0xF1, (byte) 0xC0,
	(byte) 0x43, (byte) 0x69, (byte) 0x71, (byte) 0x01, (byte) 0xAA };
    final int sizeOfLong = 8;

    // audio cover file information
    String audioFilename;
    WavFile wavFile;
    long audioFileBytesCanCover;

    // hidden file decryption information
    String passphrase;
    IDEA IDEACipher;
    CipherFeedback cipherFB;
    byte[] ivArray;
    PRPG hiddenFilePRPG; // pseudo-random permutation generator
    
    // audio cover file digest information
    byte[] audioDigestArray;

    // hidden file information
    String hiddenFileName;
    long hiddenFileLength;

    // constructor
    AudioCover(String passphrase, String audioFilename)
		    throws IOException, BadAudioFileException
    	{
	this.audioFilename = audioFilename;
	this.wavFile = new WavFile(this.audioFilename, "r");
    	this.passphrase = passphrase;
	this.IDEACipher = new IDEA(MessageDigest.hash(passphrase, new MD5()));
	this.ivArray = new byte[IDEACipher.blockLength()];
	
	// compute the audio digest for the file
	this.audioDigestArray = this.computeAudioDigest();
	this.hiddenFilePRPG = new PRPG(audioDigestArray, this.wavFile.length());

	// compute how many bytes the audio file can cover
	this.audioFileBytesCanCover = this.computeHiddenRoom();

	// see if there's an existing file in the cover
	this.hiddenFileLength = this.findHiddenFile();
    	}

    // hide a file
    public void hideFile(String fileDirectory, String fileName)
	throws IOException, BadAudioFileException
    	{
    	RandomAccessFile rf;
	byte[] inputBuffer = new byte[16384];
	int bytesRead;

	rf = new RandomAccessFile(fileDirectory + fileName, "r");
	this.startHiddenFile(fileName, rf.length());
	rf.seek(0);
	while ((bytesRead = rf.read(inputBuffer)) != 0)
	    {
	    if (bytesRead == 16384)
		this.prpgWrite(inputBuffer);
	    else
	    	{
	    	byte[] partialInputBuffer = new byte[bytesRead];
		System.arraycopy(inputBuffer, 0, partialInputBuffer, 0, bytesRead);
		this.prpgWrite(partialInputBuffer);
	    	}
	    }

	rf.close();
    	}

    // extract a hidden file
    public void saveHiddenFileAs(String fileName) throws IOException
    	{
    	FileOutputStream fo;
	byte[] outputBuffer = new byte[16384];
	int bytesRead;
	long totalBytesRead;
	boolean finishedReading = false;

	totalBytesRead = 0;
	fo = new FileOutputStream(fileName);
	do
	    {
	    if ((this.hiddenFileLength - totalBytesRead) < 16384)
	    	{
	    	outputBuffer = new byte[(int) (this.hiddenFileLength - totalBytesRead)];
		finishedReading = true;
		}
	    bytesRead = this.prpgRead(outputBuffer);
	    fo.write(outputBuffer, 0, bytesRead);
	    totalBytesRead += bytesRead;
	    } while (finishedReading == false);

	fo.close();
    	}

    // returns true if the cover has a hidden file
    public boolean hasHiddenFile()
    	{
    	if (this.hiddenFileLength > 0)
	    return  true;

	return	false;
    	}

    // returns name of hidden file
    public String getHiddenFileName()
    	{
    	return	this.hiddenFileName;
    	}

    // returns the number of bytes this audio file can cover
    public long bytesCanCover()
    	{
    	return	this.audioFileBytesCanCover;
    	}

    // close the audio cover file
    public void close () throws IOException
    	{
    	this.wavFile.close();
    	}


    // THE REST OF THE METHODS ARE THANKFULLY NOT AVAILABLE TO THE PUBLIC

    // try to create a file of the given size in the audio cover
    private void startHiddenFile (String fileName, long length) throws IOException,
	BadAudioFileException
    	{
	this.hiddenFilePRPG.reset();

	// make sure the file has room for this amount of data
    	if (this.audioFileHasRoom(fileName, length) == false)
	    throw new IOException("The file is too large too hide.");

	// close audio file and reopen for write
	this.wavFile.close();
	this.wavFile = new WavFile(this.audioFilename, "rw");

	// fill the ivArray with random numbers and write it
	Random rnd = new Random();
	for (int i = 0; i < this.ivArray.length; i++)
	    this.ivArray[i] = (byte)rnd.nextInt();
	this.cipherFB = new CipherFeedback(this.IDEACipher, this.ivArray);
	this.prpgWrite(this.ivArray);

	// write the magic signature
	byte[] encryptedMagicArray = new byte[this.magicArray.length];
	this.cipherFB.encrypt(this.magicArray, 0, encryptedMagicArray, 0,
	    this.magicArray.length);
	this.prpgWrite(encryptedMagicArray);

	// write the length of the file
	byte[] fileLengthArray = this.getArrayFromLong(length);
	this.cipherFB.encrypt(fileLengthArray);
	this.prpgWrite(fileLengthArray);

	// write the length of the filename and the filename
	byte[] hfByteArray = new byte[fileName.length() + 1];
	hfByteArray[0] = (byte) fileName.length();
	fileName.getBytes(0, fileName.length(), hfByteArray, 1);
	this.cipherFB.encrypt(hfByteArray);
	this.prpgWrite(hfByteArray);
    	}

    // compute the digest of the audio file by calculating a secure hash
    // of the passphrase and the audio data with the LSB's set to zero
    // (used by constructor function)
    private byte[] computeAudioDigest() throws IOException
    	{
	byte[] passphraseByteArray = new byte[this.passphrase.length()];
	short[] sampleBuffer = new short[16384];
	int bytesRead;
	SHA audioDigest = new SHA();

	// start at beginning of wave file and reset digest
	this.wavFile.seek(0);
	audioDigest.reset();

	// read the samples into an array of shorts and add to digest
	while ((bytesRead = this.wavFile.read(sampleBuffer)) != 0)
	    {
	    for (int i = 0; i < bytesRead; i++)
	    	{
		sampleBuffer[i] &= (short) 0xFFFE;
	    	}
	    audioDigest.add(getByteArrayFromShortArray(sampleBuffer));
	    }

	// add passphrase to digest also
	this.passphrase.getBytes(0, this.passphrase.length(), passphraseByteArray, 0);
	audioDigest.add(passphraseByteArray);

    	return audioDigest.digest();
    	}

    // looks for a hidden file in the cover, returns the length of the file if
    // one exists or 0 if not
    private long findHiddenFile() throws IOException
    	{
	long fileLength;

	this.hiddenFilePRPG.reset();

	// read iv array
	this.prpgRead(this.ivArray);
	this.cipherFB = new CipherFeedback(this.IDEACipher, this.ivArray);

	// read the magic signature and check it
	byte[] decryptedMagicArray = new byte[this.magicArray.length];
	this.prpgRead(decryptedMagicArray);
	this.cipherFB.decrypt(decryptedMagicArray);
	if (this.getLongFromArray(magicArray) !=
	    this.getLongFromArray(decryptedMagicArray))
	    return  0;

	// read the length of the file
	byte[] fileLengthArray = new byte[this.sizeOfLong];
	this.prpgRead(fileLengthArray);
	this.cipherFB.decrypt(fileLengthArray);
	fileLength = this.getLongFromArray(fileLengthArray);
	if (fileLength <= 0)
	    return  0;

	// read the filename
	byte[] hfLengthArray = new byte[1];
	this.prpgRead(hfLengthArray);
	this.cipherFB.decrypt(hfLengthArray);
	if (hfLengthArray[0] <= (byte) 0x00)
	    return  0;
	byte[] hfNameArray = new byte[(int) hfLengthArray[0]];
	this.prpgRead(hfNameArray);
	this.cipherFB.decrypt(hfNameArray);
	this.hiddenFileName = new String(hfNameArray, 0);

    	return	fileLength;
    	}

    // makes a short array into a byte array
    private byte[] getByteArrayFromShortArray (short[] shortArray)
    	{
	byte[] byteArray = new byte[2*shortArray.length];
    	
	for (int i = 0; i < shortArray.length; i++)
	    {
	    byteArray[2*i] = (byte) ((shortArray[i] >> 8) & 0xFF);
	    byteArray[2*i+1] = (byte) (shortArray[i] & 0xFF);
	    }

	return	byteArray;
    	}

    // return a long from an array of size eight
    private long getLongFromArray (byte[] byteArray)
    	{
    	long l = 0;
	long lmask;

	for (int i = 0, j = 7; i < 8; i++, j--)
	    {
	    lmask = (long) byteArray[i];
	    if (lmask < 0)
		{ lmask &= (long) 0x00FF; lmask |= (long) 0x0080; }
	    l |= (lmask << j*8);
	    }

	return	l;
    	}

    // return an array of size 8 from a long
    private byte[] getArrayFromLong (long l)
    	{
	byte[] byteArray = new byte[8];

	byteArray[0] = (byte) ((l >> 56) & 0xFF);
	byteArray[1] = (byte) ((l >> 48) & 0xFF);
	byteArray[2] = (byte) ((l >> 40) & 0xFF);
	byteArray[3] = (byte) ((l >> 32) & 0xFF);
	byteArray[4] = (byte) ((l >> 24) & 0xFF);
	byteArray[5] = (byte) ((l >> 16) & 0xFF);
	byteArray[6] = (byte) ((l >> 8) & 0xFF);
	byteArray[7] = (byte) ((l >> 0) & 0xFF);

	return	byteArray;
    	}

    // returns true if the audio file has room for hiding length amount of
    // bytes or false if not
    private boolean audioFileHasRoom(String fileName, long length)
    	{
	long bytesRequired = this.ivArray.length + this.magicArray.length +
	    this.sizeOfLong + 1 + fileName.length() + length;

    	if (bytesRequired > (this.wavFile.length() / 8) )
	    return  false;

	return	true;
    	}

    // returns the maximum number of bytes this audio file can cover
    private long computeHiddenRoom()
    	{
    	return	(this.wavFile.length() / 8) - this.ivArray.length -
	    this.magicArray.length - this.sizeOfLong - 1 - 127;
    	}

    // reads a byte array from the cover file in pseudo-random order, returns
    // number of bytes read or 0 if there are no more to be read
    private int prpgRead(byte[] byteArray) throws IOException
    	{
	int i;

	// for each byte in the array
	for (i = 0; i < byteArray.length; i++)
	    {
	    byte readByte = (byte) 0x00;

	    try 
		{
		// for each bit in this element
		for (int j = 7; j >= 0; j--)
		    {
		    this.wavFile.seek(hiddenFilePRPG.getNextIndex());
		    readByte |= (this.wavFile.getLSB() << j);
		    }
		}
	    catch (EOPException e)
	    	{
	    	return	i;
	    	}

	    byteArray[i] = readByte;
	    }

    	return	byteArray.length;
    	}

    // writes a byte array to the cover file in pseudo-random order
    private void prpgWrite(byte[] byteArray) throws IOException
    	{
	try
	    {
	    // for each byte in the array
	    for (int i = 0; i < byteArray.length; i++)
	    	{
		// for each bit in this element
		for (int j = 7; j >= 0; j--)
		    {
		    this.wavFile.seek(hiddenFilePRPG.getNextIndex());
		    this.wavFile.setLSB( (0x00 ==
			(byteArray[i] & ((byte) (0x01 << j)))) ? false : true);
		    }
	    	}	    
	    }
	catch (EOPException e)
	    {
	    throw new IOException("prpgWrite: can't write beyond end of permutation.");
	    }
    	}

    // for debugging
    // represent a byte array in ASCII hexadecimal
    private String asHex (byte[] byteArray)
    	{
	StringBuffer buf = new StringBuffer(byteArray.length * 2);

	for (int i = 0; i < byteArray.length; i++)
	    {
	    if (((int) byteArray[i] & 0xff) < 0x10)
	    buf.append("0");

	    buf.append(Long.toString((int) byteArray[i] & 0xff, 16));
	    }

	return	buf.toString();    	
    	}
    }