// StegoWav - a tool for steganography on Microsoft wave files
//
// Copyright (C) 1996 by Peter Heist <heistp@rpi.edu>.  All rights reserved.
//
// THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
// ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
// IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
// ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE
// FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
// DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
// OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
// HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
// LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
// OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
// SUCH DAMAGE.

import java.lang.*;
import java.crypt.*;

// This is the exception created when the end of the pseudo-random permutation
// is reached.
class EOPException extends Exception
    {
    public EOPException() 
    	{ super(); }
    public EOPException(String s) 
    	{ super(s); }
    }

// Pseudo-random Sequence Generator.  This class will generate a
// permutation of the set N = {1...N} (where N is the number of elements)
// whose order is based on the given byte array.  This should make it computationally
// infeasible to know the resultant permutation without knowing the
// byte array.  This is based on work originally done by Luby and Rackoff
// in 1988.  The byte array should be generated from a secure hash (such as
// SHA) from a passphrase and possibly more information.  Note that this
// algorithm is nice for composite indices, which is best suited for
// images, but here it is implemented for single index permutations.
public class PRPG 
    {
    // PRPG properties
    long elements;
    long X, Y; // composite index values
    long virtualIndex;
    long sourceIndex;

    // SHA digest
    String digestString;
    SHA prpgSHA;

    // constructor    
    PRPG (byte[] digestArray, long elements)
    	{
	this.elements = elements; // number of elements in permutation
	this.digestString = new String(digestArray, 0);
	this.prpgSHA = new SHA();
	this.sourceIndex = this.virtualIndex = 0;

	// calculate composite index values (could be better factoring algorithm)
	this.X = this.Y = (long) Math.ceil(Math.sqrt((double) elements));
    	}

    // reset PRPG to first index
    public void reset ()
    	{
    	this.sourceIndex = this.virtualIndex = 0;
	this.prpgSHA = new SHA();
    	}

    // get the next index in the permutation or throw an EOPException
    // if finished
    public long getNextIndex () throws EOPException
    	{
	long L, R, targetIndex;

	if (this.virtualIndex >= elements)
	    throw new EOPException();

	do
	    {
	    L = this.sourceIndex / X;
	    R = this.sourceIndex % X;

	    // round 1
	    L = (L + this.getPositiveLongFromByteArray(prpgSHA.hash
		(this.digestString + R))) % Y;

	    // round 2
	    R = (R + this.getPositiveLongFromByteArray(prpgSHA.hash
		(this.digestString + L))) % X;

	    // round 3
	    L = (L + this.getPositiveLongFromByteArray(prpgSHA.hash
		(this.digestString + R))) % Y;

	    this.sourceIndex++;
	    targetIndex = L * X + R; // set target index
	    } while (targetIndex >= elements);

	this.virtualIndex++;

	return	targetIndex;
    	}

    // takes the first eight bytes of a byte array (array must be >= 8 in length) and
    // puts the result into a long (and insures that the result is positive)
    // NOTE: Java has no provision for directly converting byte arrays to numbers
    //       so this or something similar has to be done
    private long getPositiveLongFromByteArray(byte[] byteArray)
    	{
	long result = 0L;

	if (byteArray.length < 8) return -1;
	
	result =
	    ( ((long) byteArray[0] & 0x7F) << 56) + // don't want negative result
	    ( ((long) byteArray[1] & 0xFF) << 48) +
	    ( ((long) byteArray[2] & 0xFF) << 40) +
	    ( ((long) byteArray[3] & 0xFF) << 32) +
	    ( ((long) byteArray[4] & 0xFF) << 24) +
	    ( ((long) byteArray[5] & 0xFF) << 16) +
	    ( ((long) byteArray[6] & 0xFF) << 8) +
	    ( ((long) byteArray[7] & 0xFF) );

	//result |= (((long) byteArray[0] & 0x7F) << 56);	// don't want negative result
	//result |= (((long) byteArray[1] & 0xFF) << 48);
	//result |= (((long) byteArray[2] & 0xFF) << 40);
	//result |= (((long) byteArray[3] & 0xFF) << 32);
	//result |= (((long) byteArray[4] & 0xFF) << 24);
	//result |= (((long) byteArray[5] & 0xFF) << 16);
	//result |= (((long) byteArray[6] & 0xFF) << 8);
	//result |= (((long) byteArray[7] & 0xFF) << 0);

	return	result;
    	}
    }