// StegoWav - a tool for steganography on Microsoft wave files
//
// Copyright (C) 1996 by Peter Heist <heistp@rpi.edu>.  All rights reserved.
//
// THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
// ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
// IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
// ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE
// FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
// DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
// OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
// HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
// LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
// OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
// SUCH DAMAGE.

import java.io.*;

// This is the bad audio file exception, which can mean the header of
// the wave file is invalid or the file is otherwise inconsistent.
class BadAudioFileException extends Exception
    {
    public BadAudioFileException() 
    	{ super(); }
    public BadAudioFileException(String s) 
    	{ super(s); }
    }


// This is the class which can perform operations on Microsoft Wave files.
// Currently it assumes the files have 16-bit samples.
public class WavFile8
    {
    // wav file constants
    final int wavFileHeaderSize = 100; // this is mainly a bogus number
				       // it should be 44 but apparently some
				       // wave files have larger headers

    // file information
    RandomAccessFile wavFile;
    long fileLength; // number of samples in audio
    int resolution;
    long samplePos; // what sample the file is positioned on
    
    // constructor function
    WavFile8 (String fileName, String mode) throws IOException, BadAudioFileException
    	{
	this.wavFile = new RandomAccessFile(fileName, mode);
	this.fileLength = (this.wavFile.length() - wavFileHeaderSize);
    	this.resolution = 8; // hardcoded to 8-bit audio
	this.samplePos = 0;
	if (this.fileLength <= 0)
	    throw new BadAudioFileException("File not long enough to be a WAV");
	this.seek(this.samplePos);
    	}

    // close the wave file
    public void close() throws IOException
    	{
    	this.wavFile.close();
    	}

    // report the length of the file IN SAMPLES
    public long length()
    	{
    	return	this.fileLength;
    	}

    // seek to a particular sample (samples indexed 0 to fileLength-1)
    public void seek(long sample) throws IOException
    	{
    	this.wavFile.seek(wavFileHeaderSize + sample);
	this.samplePos = sample;
    	}

    // set the least significant bit of the current sample to true or false
    public void setLSB(boolean b) throws IOException
    	{
	byte LSByte;

	// read least significant byte from original sample and mask appropriate bit
	LSByte = (byte) this.wavFile.read();
	if (b == true)
	    LSByte |= (byte) 0x01;
	else
	    LSByte &= (byte) 0xFE;

	// reposition and write sample back
	this.seek(this.samplePos);
	this.wavFile.write(LSByte);
	this.seek(this.samplePos);
    	}

    // get the least significant bit of the current sample
    public byte getLSB() throws IOException
    	{
	byte LSByte;

    	// read least significant byte from original sample and reposition
	LSByte = (byte) this.wavFile.read();
	this.seek(this.samplePos);
	return	(byte) (LSByte & (byte) 0x01);
    	}

    // read an array of samples, return 0 at end of file
    public int read(byte[] byteArray) throws IOException
    	{
	int bytesRead = 0;

    	if (this.samplePos >= this.length())
	    return  0;
    	
    	try 
	    {
	    bytesRead = this.wavFile.read(byteArray);
	    samplePos += bytesRead;
	    }
	catch (EOFException e)
	    {
	    
	    }

	return	bytesRead;
    	}

    // get the number of bits per sample of this file
    public int resolution()
    	{
    	return	this.resolution;
    	}
    }